"""Tests combining fork with tiledb context threads.

Background: the core tiledb library uses threads and it's easy to
experience deadlocks when forking a process that is using tiledb.  The
project doesn't have a solution for this at the moment other than to
avoid using fork(), which is the same recommendation that Python makes.
Python 3.12 warns if you fork() when multiple threads are detected and
Python 3.14 will make it so you never accidentally fork():
multiprocessing will default to "spawn" on Linux.
"""

import multiprocessing
import os
import sys
import warnings

import pytest

import tiledb


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_no_warning_fork_without_ctx():
    """Get no warning if no tiledb context exists."""
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        pid = os.fork()
        if pid == 0:
            os._exit(0)
        else:
            os.wait()


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_warning_fork_with_ctx():
    """Get a warning if we fork after creating a tiledb context."""
    _ = tiledb.Ctx()
    with pytest.warns(UserWarning, match="TileDB is a multithreading library"):
        pid = os.fork()
        if pid == 0:
            os._exit(0)
        else:
            os.wait()


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_warning_fork_with_default_ctx():
    """Get a warning if we fork after creating a default context."""
    _ = tiledb.default_ctx()
    with pytest.warns(UserWarning, match="TileDB is a multithreading library"):
        pid = os.fork()
        if pid == 0:
            os._exit(0)
        else:
            os.wait()


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
def test_no_warning_multiprocessing_without_ctx():
    """Get no warning if no tiledb context exists."""
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        mp = multiprocessing.get_context("fork")
        p = mp.Process()
        p.start()
        p.join()


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_warning_multiprocessing_with_ctx():
    """Get a warning if we fork after creating a tiledb context."""
    _ = tiledb.Ctx()
    mp = multiprocessing.get_context("fork")
    p = mp.Process()
    with pytest.warns(UserWarning, match="TileDB is a multithreading library"):
        p.start()
    p.join()


@pytest.mark.skipif(
    sys.platform == "win32", reason="fork() is not available on Windows"
)
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_warning_multiprocessing_with_default_ctx():
    """Get a warning if we fork after creating a default context."""
    _ = tiledb.default_ctx()
    mp = multiprocessing.get_context("fork")
    p = mp.Process()
    with pytest.warns(UserWarning, match="TileDB is a multithreading library"):
        p.start()
    p.join()



import io
import os
import pathlib
import pickle
import random
import sys

import numpy as np
import pytest

import tiledb

from .common import DiskTestCase, rand_utf8


class TestVFS(DiskTestCase):
    def test_supports(self):
        vfs = tiledb.VFS()

        self.assertTrue(vfs.supports("file"))
        self.assertIsInstance(vfs.supports("s3"), bool)
        self.assertIsInstance(vfs.supports("hdfs"), bool)
        self.assertIsInstance(vfs.supports("gcs"), bool)
        self.assertIsInstance(vfs.supports("azure"), bool)

        with self.assertRaises(ValueError):
            vfs.supports("invalid")

    def test_vfs_config(self):
        opt = {"region": "us-west-x1234"}
        params = [opt, tiledb.Config(opt)]
        for param in params:
            vfs = tiledb.VFS(param)
            assert vfs.config()["region"] == opt["region"]

    def test_dir(self):
        vfs = tiledb.VFS()

        dir = self.path("foo")
        self.assertFalse(vfs.is_dir(dir))

        # create
        vfs.create_dir(dir)
        if pytest.tiledb_vfs != "s3":
            self.assertTrue(vfs.is_dir(dir))

        # remove
        vfs.remove_dir(dir)
        self.assertFalse(vfs.is_dir(dir))

        # create nested path
        dir = self.path("foo/bar")
        if pytest.tiledb_vfs != "s3":
            # this fails locally because "foo" base path does not exist
            # this will not fail on s3 because there is no concept of directory
            with self.assertRaises(tiledb.TileDBError):
                vfs.create_dir(dir)

        vfs.create_dir(self.path("foo"))
        vfs.create_dir(self.path("foo/bar"))
        if pytest.tiledb_vfs != "s3":
            self.assertTrue(vfs.is_dir(dir))

    def test_file(self):
        vfs = tiledb.VFS()

        file = self.path("foo")
        self.assertFalse(vfs.is_file(file))

        # create
        vfs.touch(file)
        self.assertTrue(vfs.is_file(file))

        # remove
        vfs.remove_file(file)
        self.assertFalse(vfs.is_file(file))

        # check nested path
        file = self.path("foo/bar")
        if pytest.tiledb_vfs != "s3":
            # this fails locally because "foo" base path does not exist
            # this will not fail on s3 because there is no concept of directory
            with self.assertRaises(tiledb.TileDBError):
                vfs.touch(file)

    def test_move(self):
        vfs = tiledb.VFS()

        vfs.create_dir(self.path("foo"))
        vfs.create_dir(self.path("bar"))
        vfs.touch(self.path("bar/baz"))

        self.assertTrue(vfs.is_file(self.path("bar/baz")))

        vfs.move_file(self.path("bar/baz"), self.path("foo/baz"))

        self.assertFalse(vfs.is_file(self.path("bar/baz")))
        self.assertTrue(vfs.is_file(self.path("foo/baz")))

        # moving to invalid dir should raise an error
        if pytest.tiledb_vfs != "s3":
            # this fails locally because "foo" base path does not exist
            # this will not fail on s3 because there is no concept of directory
            with self.assertRaises(tiledb.TileDBError):
                vfs.move_dir(self.path("foo/baz"), self.path("do_not_exist/baz"))

    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="VFS copy commands from core are not supported on Windows",
    )
    def test_copy(self):
        vfs = tiledb.VFS()

        vfs.create_dir(self.path("foo"))
        vfs.create_dir(self.path("bar"))
        vfs.touch(self.path("foo/baz"))

        self.assertTrue(vfs.is_file(self.path("foo/baz")))

        vfs.copy_file(self.path("foo/baz"), self.path("bar/baz"))

        self.assertTrue(vfs.is_file(self.path("foo/baz")))
        self.assertTrue(vfs.is_file(self.path("bar/baz")))

        vfs.copy_dir(self.path("foo"), self.path("baz"))

        self.assertTrue(vfs.is_file(self.path("baz/baz")))

        # copying to invalid dir should raise an error
        if pytest.tiledb_vfs != "s3":
            # this fails locally because "foo" base path does not exist
            # this will not fail on s3 because there is no concept of directory
            with self.assertRaises(tiledb.TileDBError):
                vfs.copy_dir(self.path("foo/baz"), self.path("do_not_exist/baz"))

    def test_write_read(self):
        vfs = tiledb.VFS()

        buffer = b"bar"
        fio = vfs.open(pathlib.Path(self.path("foo")), "wb")
        fio.write(buffer)
        self.assertEqual(vfs.file_size(self.path("foo")), 3)

        fio = vfs.open(self.path("foo").encode("utf-8"), "rb")
        self.assertEqual(fio.read(3), buffer)
        # test read with numpy integers
        fio.seek(np.int64(0))
        self.assertEqual(fio.read(np.int32(3)), buffer)
        fio.seek(np.int64(0))
        self.assertEqual(fio.read(np.uint64(3)), buffer)
        fio.close()

        # write / read empty input
        fio = vfs.open(self.path("baz"), "wb")
        fio.write(b"")
        fio.close()
        self.assertEqual(vfs.file_size(self.path("baz")), 0)

        fio = vfs.open(self.path("baz"), "rb")
        self.assertEqual(fio.read(0), b"")
        fio.close()

        # read from file that does not exist
        with self.assertRaises(tiledb.TileDBError):
            vfs.open(self.path("do_not_exist"), "rb")

    def test_io(self):
        vfs = tiledb.VFS()

        buffer = b"0123456789"
        with tiledb.FileIO(vfs, self.path("foo"), mode="wb") as fio:
            fio.write(buffer)
            fio.flush()
            self.assertEqual(fio.tell(), len(buffer))

        with tiledb.FileIO(vfs, self.path("foo"), mode="rb") as fio:
            with self.assertRaises(IOError):
                fio.write(b"foo")

        self.assertEqual(vfs.file_size(self.path("foo")), len(buffer))

        fio = tiledb.FileIO(vfs, self.path("foo"), mode="rb")
        self.assertEqual(fio.read(3), b"012")
        self.assertEqual(fio.tell(), 3)
        self.assertEqual(fio.read(3), b"345")
        self.assertEqual(fio.tell(), 6)
        self.assertEqual(fio.read(10), b"6789")
        self.assertEqual(fio.tell(), 10)

        # seek from beginning
        fio.seek(0)
        self.assertEqual(fio.tell(), 0)
        self.assertEqual(fio.read(), buffer)

        # seek must be positive when SEEK_SET
        with self.assertRaises(ValueError):
            fio.seek(-1, 0)

        # seek from current positfion
        fio.seek(5)
        self.assertEqual(fio.tell(), 5)
        fio.seek(3, 1)
        self.assertEqual(fio.tell(), 8)
        fio.seek(-3, 1)
        self.assertEqual(fio.tell(), 5)

        # seek from end
        fio.seek(-4, 2)
        self.assertEqual(fio.tell(), 6)

        # Test readall
        fio.seek(0)
        self.assertEqual(fio.readall(), buffer)
        self.assertEqual(fio.tell(), 10)

        fio.seek(5)
        self.assertEqual(fio.readall(), buffer[5:])
        self.assertEqual(fio.readall(), b"")

        # Test readinto
        fio.seek(0)
        test_bytes = bytearray(10)
        self.assertEqual(fio.readinto(test_bytes), 10)
        self.assertEqual(test_bytes, buffer)

        # Reading from the end should return empty
        fio.seek(0)
        fio.read()
        self.assertEqual(fio.read(), b"")

        # Test writing and reading lines with TextIOWrapper
        lines = [rand_utf8(random.randint(0, 50)) + "\n" for _ in range(10)]
        rand_uri = self.path("test_fio.rand")
        with tiledb.FileIO(vfs, rand_uri, "wb") as f:
            txtio = io.TextIOWrapper(f, encoding="utf-8")
            txtio.writelines(lines)
            txtio.flush()

        with tiledb.FileIO(vfs, rand_uri, "rb") as f2:
            txtio = io.TextIOWrapper(f2, encoding="utf-8")
            self.assertEqual(txtio.readlines(), lines)

    def test_pickle(self):
        # test that vfs can be pickled and unpickled with config options
        config = tiledb.Config(
            {"vfs.s3.region": "eu-west-1", "vfs.max_parallel_ops": "1"}
        )
        vfs = tiledb.VFS(config)
        with io.BytesIO() as buf:
            pickle.dump(vfs, buf)
            buf.seek(0)
            vfs2 = pickle.load(buf)

            self.assertIsInstance(vfs2, tiledb.VFS)
            self.assertEqual(vfs2.config()["vfs.s3.region"], "eu-west-1")
            self.assertEqual(vfs2.config()["vfs.max_parallel_ops"], "1")

    def test_sc42569_vfs_memoryview(self):
        # This test is to ensure that giving np.ndarray buffer to readinto works
        # when trying to write bytes that cannot be converted to float32 or int32
        vfs = tiledb.VFS()

        buffer = b"012\x00\x01"
        with tiledb.FileIO(vfs, self.path("foo"), mode="wb") as fio:
            fio.write(buffer)
            fio.flush()
            self.assertEqual(fio.tell(), len(buffer))

        fio = tiledb.FileIO(vfs, self.path("foo"), mode="rb")

        # Test readinto with np.float32
        fio.seek(0)
        test_np_array = np.empty(5, dtype=np.float32)
        n_bytes = fio.readinto(test_np_array)
        self.assertEqual(n_bytes, 5)
        self.assertEqual(test_np_array.tobytes()[:n_bytes], buffer)

        # Test readinto with np.int32
        fio.seek(0)
        test_np_array = np.empty(5, dtype=np.int32)
        n_bytes = fio.readinto(test_np_array)
        self.assertEqual(n_bytes, 5)
        self.assertEqual(test_np_array.tobytes()[:n_bytes], buffer)

    def test_ls(self):
        basepath = self.path("test_vfs_ls")
        self.vfs.create_dir(basepath)
        for id in (1, 2, 3):
            dir = os.path.join(basepath, f"dir{id}")
            self.vfs.create_dir(dir)
            fname = os.path.join(basepath, "file_" + str(id))
            with tiledb.FileIO(self.vfs, fname, "wb") as fio:
                fio.write(b"")

        expected = ("file_1", "file_2", "file_3")
        # empty directories do not "exist" on s3
        if pytest.tiledb_vfs != "s3":
            expected = expected + ("dir1", "dir2", "dir3")

        self.assertSetEqual(
            set(expected),
            set(
                map(
                    lambda x: os.path.basename(x.split("test_vfs_ls")[1]),
                    self.vfs.ls(basepath),
                )
            ),
        )

    @pytest.mark.skipif(
        pytest.tiledb_vfs not in ["file", "s3", "azure", "gcs"],
        reason="Only test on local, S3, Azure, and GCS",
    )
    def test_ls_recursive(self):
        # Create a nested directory structure to test recursive listing
        basepath = self.path("test_vfs_ls_recursive")
        self.vfs.create_dir(basepath)

        dir = os.path.join(basepath, "dir1")
        self.vfs.create_dir(dir)

        fname = os.path.join(dir, "file_1")
        with tiledb.FileIO(self.vfs, fname, "wb") as fio:
            fio.write(b"")

        fname = os.path.join(dir, "file_2")
        with tiledb.FileIO(self.vfs, fname, "wb") as fio:
            fio.write(b"")

        dir = os.path.join(basepath, "dir2")
        self.vfs.create_dir(dir)

        dir2 = os.path.join(dir, "dir2_1")
        self.vfs.create_dir(dir2)

        fname = os.path.join(dir2, "file_1")
        with tiledb.FileIO(self.vfs, fname, "wb") as fio:
            fio.write(b"")
        fname = os.path.join(dir2, "file_2")
        with tiledb.FileIO(self.vfs, fname, "wb") as fio:
            fio.write(b"")

        dir2 = os.path.join(dir, "dir2_2")
        self.vfs.create_dir(dir2)

        fname = os.path.join(dir2, "file_1")
        with tiledb.FileIO(self.vfs, fname, "wb") as fio:
            fio.write(b"")

        expected = [
            "dir1",
            "dir1/file_1",
            "dir1/file_2",
            "dir2",
            "dir2/dir2_1",
            "dir2/dir2_1/file_1",
            "dir2/dir2_1/file_2",
            "dir2/dir2_2",
            "dir2/dir2_2/file_1",
        ]

        self.assertSetEqual(
            set(expected),
            set(
                map(
                    # # Keep only the paths after the basepath and normalize them to work on all platforms
                    lambda x: os.path.normpath(
                        x.split("test_vfs_ls_recursive/")[1]
                    ).replace("\\", "/"),
                    self.vfs.ls_recursive(basepath),
                )
            ),
        )

        # Check with user provided callback
        callback_results = []

        def callback(uri, _):  # we don't use the second argument 'is_dir'
            callback_results.append(uri)
            return True

        self.vfs.ls_recursive(basepath, callback)

        self.assertSetEqual(
            set(expected),
            set(
                map(
                    # Keep only the paths after the basepath and normalize them to work on all platforms
                    lambda x: os.path.normpath(
                        x.split("test_vfs_ls_recursive/")[1]
                    ).replace("\\", "/"),
                    callback_results,
                )
            ),
        )

        # Can also be called by calling ls with recursive=True
        self.assertSetEqual(
            set(expected),
            set(
                map(
                    # Keep only the paths after the basepath and normalize them to work on all platforms
                    lambda x: os.path.normpath(
                        x.split("test_vfs_ls_recursive/")[1]
                    ).replace("\\", "/"),
                    self.vfs.ls(basepath, recursive=True),
                )
            ),
        )

    def test_dir_size(self):
        vfs = tiledb.VFS()

        path = self.path("test_vfs_dir_size")
        vfs.create_dir(path)
        rand_sizes = np.random.choice(100, size=4, replace=False)
        for size in rand_sizes:
            file_path = os.path.join(path, "f_" + str(size))
            with tiledb.FileIO(vfs, file_path, "wb") as f:
                data = os.urandom(size)
                f.write(data)

        self.assertEqual(vfs.dir_size(path), sum(rand_sizes))

    def test_open_with(self):
        uri = self.path("test_open_with")
        vfs = tiledb.VFS()
        buffer = b"0123456789"

        with vfs.open(uri, mode="wb") as fio:
            fio.write(buffer)
            fio.flush()
            self.assertEqual(fio.tell(), len(buffer))

        with vfs.open(uri, mode="rb") as fio:
            with self.assertRaises(IOError):
                fio.write(b"foo")
            self.assertEqual(fio.read(len(buffer)), buffer)


def test_vfs_isdir(tmp_path):
    """isdir is an alias for is_dir."""
    fs = tiledb.VFS()
    assert fs.isdir(tmp_path.as_posix())


def test_vfs_isfile(tmp_path):
    """isfile is an alias for is_file."""
    tmp_file = tmp_path.joinpath("foo")
    tmp_file.touch()
    fs = tiledb.VFS()
    assert fs.isfile(tmp_file.as_posix())


def test_vfs_size(tmp_path):
    """size is an alias for file_size."""
    tmp_file = tmp_path.joinpath("foo")
    buffer = b"0123456789"
    tmp_file.write_bytes(buffer)
    fs = tiledb.VFS()
    assert fs.size(tmp_file.as_posix()) == len(buffer)



import copy
import random

import numpy as np
from numpy.testing import assert_array_equal

import tiledb
import tiledb.main as core

from .common import DiskTestCase, rand_ascii


class CoreCCTest(DiskTestCase):
    def test_pyquery_basic(self):
        ctx = tiledb.Ctx()
        uri = self.path("test_pyquery_basic")
        with tiledb.from_numpy(uri, np.random.rand(4)):
            pass

        with tiledb.open(uri) as a:
            with tiledb.scope_ctx({"py.init_buffer_bytes": "abcd"}) as testctx:
                with self.assertRaises(ValueError):
                    core.PyQuery(testctx, a, ("",), (), 0, False)

            q = core.PyQuery(ctx, a, ("",), (), 0, False)

            try:
                q._test_err("bad foo happened")
            except Exception as exc:
                assert isinstance(exc, tiledb.TileDBError)
                assert str(exc) == "bad foo happened"

        with tiledb.open(uri) as a:
            q2 = core.PyQuery(ctx, a, ("",), (), 0, False)
            subarray = tiledb.Subarray(a)
            subarray.add_ranges([[(0, 3)]])
            q2.set_subarray(subarray)
            q2.submit()
            res = q2.results()[""][0]
            res.dtype = np.double
            assert_array_equal(res, a[:])

    def test_pyquery_init(self):
        uri = self.path("test_pyquery_init")
        intmax = np.iinfo(np.int64).max
        config_dict = {
            "sm.tile_cache_size": "100",
            "py.init_buffer_bytes": str(intmax),
            "py.alloc_max_bytes": str(intmax),
        }
        with tiledb.scope_ctx(config_dict) as ctx:
            with tiledb.from_numpy(uri, np.random.rand(4)):
                pass

            with tiledb.open(uri) as a:
                q = core.PyQuery(ctx, a, ("",), (), 0, False)
                self.assertEqual(q._test_init_buffer_bytes, intmax)
                self.assertEqual(q._test_alloc_max_bytes, intmax)

                with self.assertRaisesRegex(
                    ValueError,
                    "Invalid parameter: 'py.alloc_max_bytes' must be >= 1 MB ",
                ), tiledb.scope_ctx({"py.alloc_max_bytes": 10}) as ctx2:
                    q = core.PyQuery(ctx2, a, ("",), (), 0, False)

    def test_import_buffer(self):
        uri = self.path("test_import_buffer")

        def_tile = 1
        if tiledb.libtiledb.version() < (2, 2):
            def_tile = 2

        dom = tiledb.Domain(
            tiledb.Dim(domain=(0, 3), tile=def_tile, dtype=np.int64),
            tiledb.Dim(domain=(0, 3), tile=def_tile, dtype=np.int64),
        )
        attrs = [
            tiledb.Attr(name="", dtype=np.float64),
            tiledb.Attr(name="foo", dtype=np.int32),
            tiledb.Attr(name="str", dtype=str),
        ]
        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=False)
        tiledb.DenseArray.create(uri, schema)

        data_orig = {
            "": 2.5 * np.identity(4, dtype=np.float64),
            "foo": 8 * np.identity(4, dtype=np.int32),
            "str": np.array(
                [rand_ascii(random.randint(0, 5)) for _ in range(16)], dtype="U0"
            ).reshape(4, 4),
        }

        with tiledb.open(uri, "w") as A:
            A[:] = data_orig

        with tiledb.open(uri) as B:
            assert_array_equal(B[:][""], data_orig[""]),
            assert_array_equal(B[:]["foo"], data_orig["foo"])

        data_mod = {
            "": 5 * np.identity(4, dtype=np.float64),
            "foo": 32 * np.identity(4, dtype=np.int32),
            "str": np.array(
                [rand_ascii(random.randint(1, 7)) for _ in range(16)], dtype="U0"
            ).reshape(4, 4),
        }

        str_offsets = np.array(
            [0] + [len(x) for x in data_mod["str"].flatten()[:-1]], dtype=np.uint64
        )
        str_offsets = np.cumsum(str_offsets)

        str_raw = np.array(
            [ord(c) for c in "".join([x for x in data_mod["str"].flatten()])],
            dtype=np.uint8,
        )

        data_mod_bfr = {
            "": (data_mod[""].flatten().view(np.uint8), np.array([], dtype=np.uint64)),
            "foo": (
                data_mod["foo"].flatten().view(np.uint8),
                np.array([], dtype=np.uint64),
            ),
            "str": (str_raw.flatten().view(np.uint8), str_offsets),
        }

        with tiledb.open(uri) as C:
            res = C.multi_index[0:3, 0:3]
            assert_array_equal(res[""], data_orig[""])
            assert_array_equal(res["foo"], data_orig["foo"])
            assert_array_equal(res["str"], data_orig["str"])

            C._set_buffers(copy.deepcopy(data_mod_bfr))
            res = C.multi_index[0:3, 0:3]
            assert_array_equal(res[""], data_mod[""])
            assert_array_equal(res["foo"], data_mod["foo"])
            assert_array_equal(res["str"], data_mod["str"])

        with tiledb.open(uri) as D:
            D._set_buffers(copy.deepcopy(data_mod_bfr))
            res = D[:, :]
            assert_array_equal(res[""], data_mod[""])
            assert_array_equal(res["foo"], data_mod["foo"])
            assert_array_equal(res["str"], data_mod["str"])

        with tiledb.DenseArray(uri, mode="r") as E, tiledb.scope_ctx() as ctx:
            # Ensure that query only returns specified attributes
            q = core.PyQuery(ctx, E, ("foo",), (), 0, False)
            subarray = tiledb.Subarray(E, ctx)
            subarray.add_ranges([[(0, 1)]])
            q.set_subarray(subarray)
            q.submit()
            r = q.results()
            self.assertTrue("foo" in r)
            self.assertTrue("str" not in r)
            del q



import tempfile

import numpy as np

import tiledb


class MetadataTest:
    def setup(self):
        self.path = tempfile.mkdtemp()
        print(self.path)
        self.array = np.random.rand(4)
        tiledb.from_numpy(self.path, self.array)


class MetadataWrite(MetadataTest):
    def setup(self):
        super().setup()

    def time_write(self):
        with tiledb.open(self.path, "w") as A:
            for i in range(1_000_000):
                A.meta["x"] = "xyz"


class MetadataRead(MetadataTest):
    def setup(self):
        super().setup()

        with tiledb.open(self.path, "w") as A:
            A.meta["x"] = "xyz"

    def time_read(self):
        with tiledb.open(self.path) as A:
            for i in range(1_000_000):
                A.meta["x"]



import io
import warnings
from typing import Any, Optional, Sequence, Union

import numpy as np

import tiledb.libtiledb as lt

from .ctx import Ctx, CtxMixin
from .datatypes import DataType
from .filter import Filter, FilterList


class Attr(CtxMixin, lt.Attribute):
    """
    Represents a TileDB attribute.
    """

    def __init__(
        self,
        name: str = "",
        dtype: np.dtype = np.float64,
        fill: Any = None,
        var: bool = None,
        nullable: bool = False,
        filters: Union[FilterList, Sequence[Filter]] = None,
        enum_label: str = None,
        ctx: Optional[Ctx] = None,
    ):
        """Class representing a TileDB array attribute.

        :param name: Attribute name, empty if anonymous
        :param dtype: Attribute value datatype
        :param fill: Fill value for unset cells
        :param var: Attribute is variable-length (automatic for byte/string types)
        :param nullable: Attribute is nullable
        :param filters: List of filters to apply
        :param ctx: A TileDB Context
        :raises TypeError: invalid dtype
        :raises tiledb.TileDBError:
        """
        dt = DataType.from_numpy(
            np.dtype(dtype) if dtype not in ("ascii", "blob", "wkb", "wkt") else dtype
        )

        # ensure that all strings are var-length
        if (var is None and dtype == "ascii") or np.issubdtype(dt.np_dtype, np.str_):
            var = True
        elif np.issubdtype(dt.np_dtype, np.bytes_):
            if dt.np_dtype.itemsize > 0 and var:
                warnings.warn(
                    f"Attr given `var=True` but `dtype` `{dtype}` is fixed; "
                    "setting `dtype=S0`. Hint: set `var=True` with `dtype=S0`, "
                    f"or `var=False`with `dtype={dtype}`",
                    DeprecationWarning,
                )
            elif dt.np_dtype.itemsize == 0 and dtype != "ascii":
                if var is False:
                    warnings.warn(
                        "Attr given `var=False` but `dtype` `S0` is var-length; "
                        "setting `var=True` and `dtype=S0`. Hint: set `var=False` "
                        "with `dtype=S0`, or `var=False` with a fixed-width "
                        "string `dtype=S<n>` where is  n>1",
                        DeprecationWarning,
                    )
                var = True

        super().__init__(ctx, name, dt.tiledb_type)

        if var:
            self._ncell = lt.TILEDB_VAR_NUM()
        elif dt.ncells != lt.TILEDB_VAR_NUM():
            self._ncell = dt.ncells
        else:
            raise TypeError("dtype is not compatible with var-length attribute")

        if filters is not None:
            if isinstance(filters, FilterList):
                self._filters = filters
            else:
                self._filters = FilterList(filters)

        if fill is not None:
            if self._tiledb_dtype == lt.DataType.STRING_UTF8:
                self._fill = np.array([fill.encode("utf-8")], dtype="S")
            elif self.dtype == np.dtype("complex64") or self.dtype == np.dtype(
                "complex128"
            ):
                if hasattr(fill, "dtype") and fill.dtype in {
                    np.dtype("f4, f4"),
                    np.dtype("f8, f8"),
                }:
                    _fill = fill["f0"] + fill["f1"] * 1j
                elif hasattr(fill, "__len__") and len(fill) == 2:
                    _fill = fill[0] + fill[1] * 1j
                else:
                    _fill = fill
                self._fill = np.array(_fill, dtype=self.dtype)
            else:
                self._fill = np.array([fill], dtype=self.dtype)

        if nullable is not None:
            self._nullable = nullable

        if enum_label is not None:
            self._set_enumeration_name(self._ctx, enum_label)

    def __eq__(self, other):
        if not isinstance(other, Attr):
            return False
        if self.isnullable != other.isnullable or self.dtype != other.dtype:
            return False
        if not self.isnullable:
            # Check the fill values are equal.
            def equal_or_nan(x, y):
                return x == y or (np.isnan(x) and np.isnan(y))

            if self.ncells == 1:
                if not equal_or_nan(self.fill, other.fill):
                    return False
            elif np.issubdtype(self.dtype, np.bytes_) or np.issubdtype(
                self.dtype, np.str_
            ):
                if self.fill != other.fill:
                    return False
            elif self.dtype in {np.dtype("complex64"), np.dtype("complex128")}:
                if not (
                    equal_or_nan(np.real(self.fill), np.real(other.fill))
                    and equal_or_nan(np.imag(self.fill), np.imag(other.fill))
                ):
                    return False
            else:
                if not all(
                    equal_or_nan(x, y)
                    or (
                        isinstance(x, str)
                        and x.lower() == "nat"
                        and isinstance(y, str)
                        and y.lower() == "nat"
                    )
                    for x, y in zip(self.fill[0], other.fill[0])
                ):
                    return False
        return (
            self._internal_name == other._internal_name
            and self.isvar == other.isvar
            and self.filters == other.filters
        )

    def dump(self):
        """Dumps a string representation of the Attr object to standard output (stdout)"""
        print(self._dump(), "\n")

    @property
    def dtype(self) -> np.dtype:
        """Return numpy dtype object representing the Attr type

        :rtype: numpy.dtype

        """
        return DataType.from_tiledb(self._tiledb_dtype, self._ncell).np_dtype

    @property
    def name(self) -> str:
        """Attribute string name, empty string if the attribute is anonymous

        :rtype: str
        :raises: :py:exc:`tiledb.TileDBError`

        """
        internal_name = self._name
        # handle __attr names from arrays written with libtiledb < 2
        if internal_name == "__attr":
            return ""
        return internal_name

    @property
    def _internal_name(self):
        return self._name

    @property
    def isanon(self) -> bool:
        """True if attribute is an anonymous attribute

        :rtype: bool

        """
        return self._name == "" or self._name.startswith("__attr")

    @property
    def filters(self) -> FilterList:
        """FilterList of the TileDB attribute

        :rtype: tiledb.FilterList
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return FilterList.from_pybind11(self._ctx, self._filters)

    @property
    def fill(self) -> Any:
        """Fill value for unset cells of this attribute

        :rtype: depends on dtype
        :raises: :py:exc:`tiledb.TileDBERror`
        """
        dtype = self.dtype
        if np.issubdtype(dtype, np.bytes_):
            return self._fill.tobytes()
        if np.issubdtype(dtype, np.str_):
            return self._fill.tobytes().decode("utf-8")
        if np.issubdtype(dtype, np.datetime64):
            return self._fill[0].astype(np.timedelta64)
        return self._fill

    @property
    def isnullable(self) -> bool:
        """True if the attribute is nullable

        :rtype: bool
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self._nullable

    @property
    def isvar(self) -> bool:
        """True if the attribute is variable length

        :rtype: bool
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self._var

    @property
    def ncells(self) -> int:
        """The number of cells (scalar values) for a given attribute value

        :rtype: int
        :raises: :py:exc:`tiledb.TileDBError`

        """
        assert self._ncell != 0
        return int(self._ncell)

    @property
    def isascii(self) -> bool:
        """True if the attribute is TileDB dtype TILEDB_STRING_ASCII

        :rtype: bool
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self._tiledb_dtype == lt.DataType.STRING_ASCII

    @property
    def enum_label(self):
        return self._get_enumeration_name(self._ctx)

    def __repr__(self):
        # use safe repr if pybind11 constructor failed
        if self._ctx is None:
            return object.__repr__(self)

        filters_str = ""
        if self.filters:
            filters_str = ", filters=FilterList(["
            for f in self.filters:
                filters_str += repr(f) + ", "
            filters_str += "])"

        if self._tiledb_dtype == lt.DataType.STRING_ASCII:
            attr_dtype = "ascii"
        elif self._tiledb_dtype == lt.DataType.BLOB:
            attr_dtype = "blob"
        elif (
            hasattr(lt.DataType, "GEOM_WKB")
            and self._tiledb_dtype == lt.DataType.GEOM_WKB
        ):
            attr_dtype = "wkb"
        elif (
            hasattr(lt.DataType, "GEOM_WKT")
            and self._tiledb_dtype == lt.DataType.GEOM_WKT
        ):
            attr_dtype = "wkt"
        else:
            attr_dtype = self.dtype

        if self.enum_label is None:
            enum_label = None
        else:
            enum_label = f"'{self.enum_label!s}'"

        # filters_str must be last with no spaces
        return (
            f"""Attr(name={repr(self.name)}, dtype='{attr_dtype!s}', """
            f"""var={self.isvar!s}, nullable={self.isnullable!s}, """
            f"""enum_label={enum_label}{filters_str})"""
        )

    def _repr_html_(self):
        output = io.StringIO()

        output.write("<table>")
        output.write("<tr>")
        output.write("<th>Name</th>")
        output.write("<th>Data Type</th>")
        output.write("<th>Is Var-Len</th>")
        output.write("<th>Is Nullable</th>")
        output.write("<th>Filters</th>")
        output.write("</tr>")
        output.write(f"{self._repr_html_row_only_()}")
        output.write("</table>")

        return output.getvalue()

    def _repr_html_row_only_(self):
        output = io.StringIO()

        output.write("<tr>")
        output.write(f"<td>{self.name}</td>")
        output.write(f"<td>{'ascii' if self.isascii else self.dtype}</td>")
        output.write(f"<td>{self.isvar}</td>")
        output.write(f"<td>{self.isnullable}</td>")
        output.write(f"<td>{self.filters._repr_html_()}</td>")
        output.write("</tr>")

        return output.getvalue()
